//
// <copyright file="AlbumPhotoStorage.cs" company="Microsoft">
//     Copyright (c) Microsoft Corporation.  All rights reserved.
// </copyright>
//
using System;
using System.Configuration;
using System.Web.Security;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Configuration.Provider;
using System.Web.Hosting;
using System.Xml;
using System.Security.Permissions;
using System.Net;
using System.Web;
using System.IO;
using System.Linq;
using System.Diagnostics;
using System.Text.RegularExpressions;
using System.Data.Services.Client;
using ServiceRuntime = Microsoft.ServiceHosting.ServiceRuntime;
using Microsoft.Samples.ServiceHosting.StorageClient;


namespace Microsoft.Samples.ServiceHosting.ClouderaCMS
{
    /// <summary>
    /// This class allows DevtableGen to generate the correct tables
    /// (named 'Albums' and 'Photos')
    /// </summary>

    internal class PersonalWebSiteContext : TableStorageDataServiceContext
    {
        public IQueryable<AlbumRow> Albums
        {
            get
            {
                return this.CreateQuery<AlbumRow>("Albums");
            }
        }

        public IQueryable<PhotoRow> Photos
        {
            get
            {
                return this.CreateQuery<PhotoRow>("Photos");
            }
        }
    }

    public class AlbumRow : TableStorageEntity
    {

        // AlbumId is partitionKey
        // rowKey is always an empty string
        public AlbumRow(string partitionKey) {
            SecUtility.CheckParameter(ref partitionKey, true, true, true, TableStorageConstants.MaxStringPropertySizeInChars, "partitionKey");

            PartitionKey = partitionKey;
            RowKey = string.Empty;
            Caption = string.Empty;
            IsPublic = false;
        }

        public AlbumRow()
            : base()
        {
        }

        public override string PartitionKey
        {
            set
            {
                AlbumId = value;
            }
            get
            {
                return AlbumId;
            }
        }

        public override string RowKey
        {
            get
            {
                return base.RowKey;
            }
            set
            {
                base.RowKey = value;
            }
        }

        public string AlbumId
        {
            set;
            get;
        }

        public string Caption
        {
            set;
            get;
        }

        public bool IsPublic
        {
            set;
            get;
        }

    }


    /// <summary>
    /// This class represents a row in the Photo table in the table storage.
    /// </summary>
    public class PhotoRow : TableStorageEntity
    {

        // AlbumId is partitionKey
        // PhotoId is rowKey
        public PhotoRow(string partitionKey, string rowKey)
        {
            SecUtility.CheckParameter(ref partitionKey, true, true, true, TableStorageConstants.MaxStringPropertySizeInChars, "partitionKey");
            SecUtility.CheckParameter(ref rowKey, true, true, true, TableStorageConstants.MaxStringPropertySizeInChars, "rowKey");

            PartitionKey = partitionKey;
            RowKey = rowKey;
            Caption = string.Empty;
            OriginalPhotoBlobName = string.Empty;
            SmallPhotoBlobName = string.Empty;
            MediumPhotoBlobName = string.Empty;
            LargePhotoBlobName = string.Empty;
        }

        public PhotoRow()
            : base()
        {
        }

        public override string PartitionKey
        {
            set
            {
                AlbumId = value;
            }
            get
            {
                return AlbumId;
            }
        }

        public override string RowKey
        {
            set
            {
                PhotoId = value;
            }
            get
            {
                return PhotoId;
            }
        }

        public string AlbumId
        {
            set;
            get;
        }

        public string PhotoId
        {
            set;
            get;
        }

        public string Caption
        {
            set;
            get;
        }

        public string OriginalPhotoBlobName
        {
            set;
            get;
        }

        public string SmallPhotoBlobName
        {
            set;
            get;
        }

        public string MediumPhotoBlobName
        {
            set;
            get;
        }

        public string LargePhotoBlobName
        {
            set;
            get;
        }
    }


    /// <summary>
    /// This class represents the Album table in the table storage.
    /// </summary>
    public class AlbumTable 
    {

        #region Member variables and constants

        private static string _tableName;

        private static StorageAccountInfo _account;
        private static TableStorage _tableStorage;

        private const int NumRetries = 3;
        private static RetryPolicy _tableRetry = ProviderRetryPolicies.RetryN(NumRetries, TimeSpan.FromSeconds(1));

        #endregion


        #region Public methods

        /// <summary>
        /// This is the initialization function that initializes the member
        /// variables. It also creates the Album table in the table storage
        /// if the table has not been created. This function is only called 
        /// once in the Application_Start function in Global.asax file.
        /// </summary>
        public static void Initialize()
        {
            _tableName = ConfigurationSettings.AppSettings["albumTableName"];

            try
            {
                _account = StorageAccountInfo.GetDefaultTableStorageAccountFromConfiguration();
                _tableStorage = TableStorage.Create(_account);

                _tableStorage.TryCreateTable(_tableName);
            }
            catch (Exception e)
            {
                Log.Write(EventKind.Error, "Initialization of data service structures for album storage failed!");
                throw new Exception("Could not create or find album table in the data service ", e);
            }
        }


        /// <summary>
        /// Add a row in the Album table.
        /// </summary>
        /// <param name="caption">The caption of an album </param>
        /// <param name="isPublic">If isPublic is set to true, an album can be
        /// viewed by any user. Otherwise this album can only be viewed by the
        /// users with admin privilege.
        /// </param>
        public static void AddAlbum(string caption, bool isPublic)
        {
            SecUtility.CheckParameter(ref caption, false, false, false, TableStorageConstants.MaxStringPropertySizeInChars, "caption");
            string albumId = Guid.NewGuid().ToString();

            AlbumRow newAlbumRow = new AlbumRow(albumId);
            newAlbumRow.Caption = caption;
            newAlbumRow.IsPublic = isPublic;
            
            try
            {
                TableStorageDataServiceContext svc = CreateDataServiceContext();
                svc.AddObject(_tableName, newAlbumRow);
                svc.SaveChanges();
            }
            catch (InvalidOperationException e)
            {
                HttpStatusCode status;
                // when retry policies are used we cannot distinguish between a conflict and success
                // so, in the case of a conflict, we just retrun success here
                if (TableStorageHelpers.EvaluateException(e, out status) && status == HttpStatusCode.Conflict)
                {
                    return;
                }

                Log.Write(EventKind.Warning, "Could not add the album (ID={0}) to the Album table.", albumId);
                throw new Exception("Failed to access the Album table ", e);
            }
        }


        /// <summary>
        /// Return all the albums in the Album table.
        /// </summary>
        /// <returns></returns>
        public static List<AlbumRow> GetAllAlbums()
        {
            TableStorageDataServiceContext svc = CreateDataServiceContext();

            IEnumerable<AlbumRow> query = from c in svc.CreateQuery<AlbumRow>(_tableName)
                                          select c;

            TableStorageDataServiceQuery<AlbumRow> q = new TableStorageDataServiceQuery<AlbumRow>(query as DataServiceQuery<AlbumRow>, _tableRetry);
            IEnumerable<AlbumRow> albumRows = q.ExecuteAllWithRetries();

            if (albumRows != null)
            {
                List<AlbumRow> list = new List<AlbumRow>(albumRows);
                return list;
            }

            return null;
        }


        /// <summary>
        /// Return all the albums which are public and contain at least one photo.
        /// </summary>
        /// <returns></returns>
        public static List<AlbumRow> GetNonEmptyPublicAlbums()
        {
            List<AlbumRow> nonEmptyPublicAlbums = new List<AlbumRow>(); 

            List<AlbumRow> albums = GetAllAlbums();
            foreach (AlbumRow album in albums)
            {
                if (album.IsPublic)
                {
                    List<PhotoRow> photos = PhotoTable.GetPhotos(album.AlbumId);
                    if (photos.Count > 0)
                    {
                        nonEmptyPublicAlbums.Add(album);
                    }
                }
            }

            return nonEmptyPublicAlbums;
        }


        /// <summary>
        /// Remove a row in the Album table.
        /// </summary>
        /// <param name="albumId">The album Id</param>
        public static void RemoveAlbum(string albumId)
        {
            try
            {
                TableStorageDataServiceContext svc = CreateDataServiceContext();

                IEnumerable<AlbumRow> query = from c in svc.CreateQuery<AlbumRow>(_tableName)
                                              where c.PartitionKey == albumId
                                              select c;

                TableStorageDataServiceQuery<AlbumRow> q = new TableStorageDataServiceQuery<AlbumRow>(query as DataServiceQuery<AlbumRow>, _tableRetry);
                IEnumerable<AlbumRow> albums = q.ExecuteAllWithRetries();

                int count = 0;
                foreach (AlbumRow album in albums)
                {
                    if (count++ > 1)
                    {
                        Log.Write(EventKind.Warning, "There are multiple albums with the same ID {0} in the Album table. Album ID should be unique.", albumId);
                    }

                    PhotoTable.RemovePhotos(albumId);

                    svc.DeleteObject(album);
                    svc.SaveChanges();
                }
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "The deletion of the album (ID={0}) from the data storage failed.", albumId);
                throw new Exception("Failed to delete the album ", e);
            }
        }


        /// <summary>
        /// Modify the caption and isPublic columns in the Album table.
        /// </summary>
        /// <param name="caption">The caption of an album</param>
        /// <param name="isPublic">If isPublic is set to true, an album can be
        /// viewed by any user. Otherwise this album can only be viewed by the
        /// users with admin privilege.
        /// </param>
        /// <param name="albumId">The album Id</param>
        public static void EditAlbum(string caption, bool isPublic, string albumId)
        {
            SecUtility.CheckParameter(ref caption, false, false, false, TableStorageConstants.MaxStringPropertySizeInChars, "caption");

            try
            {
                TableStorageDataServiceContext svc = CreateDataServiceContext();

                IEnumerable<AlbumRow> query = from c in svc.CreateQuery<AlbumRow>(_tableName)
                                              where c.PartitionKey == albumId
                                              select c;

                TableStorageDataServiceQuery<AlbumRow> q = new TableStorageDataServiceQuery<AlbumRow>(query as DataServiceQuery<AlbumRow>, _tableRetry);
                IEnumerable<AlbumRow> albums = q.ExecuteAllWithRetries();

                int count = 0;
                foreach (AlbumRow album in albums)
                {
                    if (count++ > 1)
                    {
                        Log.Write(EventKind.Warning, "There are multiple albums with the same ID (ID={0}) in the Album table. Album ID should be unique.", albumId);
                    }
                    album.Caption = caption;
                    album.IsPublic = isPublic;
                    svc.UpdateObject(album);
                    svc.SaveChanges();
                }
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "Could not edit the album (ID={0}) in the Album table.", albumId);
                throw new Exception("Failed to access the Album table ", e);
            }
        }

        public static AlbumRow GetAlbum(string albumId)
        {
            TableStorageDataServiceContext svc = CreateDataServiceContext(); 

            return GetAlbum(albumId, svc);
        }

        #endregion

        #region Help methods
        private static TableStorageDataServiceContext CreateDataServiceContext()
        {
            return _tableStorage.GetDataServiceContext();
        }

        private static AlbumRow GetAlbum(string albumId, DataServiceContext svc)
        {
            Debug.Assert(svc != null);

            try
            {
                IEnumerable<AlbumRow> query = from c in svc.CreateQuery<AlbumRow>(_tableName)
                                              where c.PartitionKey == albumId
                                              select c;

                TableStorageDataServiceQuery<AlbumRow> q = new TableStorageDataServiceQuery<AlbumRow>(query as DataServiceQuery<AlbumRow>, _tableRetry);
                IEnumerable<AlbumRow> albums = q.ExecuteAllWithRetries();

                int count = 0;
                foreach (AlbumRow album in albums)
                {
                    if (count++ > 1)
                    {
                        Log.Write(EventKind.Warning, "There are multiple albums with the same ID (ID={0}) in the Album table. Album ID should be unique.", albumId);
                    }
                    return album;
                }

                return null;
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "Could not retrieve the album (ID={0}) from the Album table.", albumId);
                throw new Exception("Failed to access the Album table ", e);
            }
        }

        #endregion
    }


    /// <summary>
    /// This class represents the Photo table in the table storage. The Photo
    /// table contains the metadata of a photo, such as the caption of a 
    /// photo and the names of the four different sized copies of this photo 
    /// (original-sized, large-sized, medium-sized and small-sized copies) that 
    /// are storaged in the blob storage.
    /// </summary>
    public class PhotoTable
    {

        #region Member variables and constants

        private static string _accountName;
        private static string _accountSharedKey;
        private static string _storageEndpoint;
        private static bool _usePathStyleUris;

        private static string _tableName;

        private static string _largePhotoBlobName;
        private static string _mediumPhotoBlobName;
        private static string _smallPhotoBlobName;
        private static string _originalPhotoBlobName;

        private static StorageAccountInfo _account;
        private static TableStorage _tableStorage;

        private static BlobStorage _blobStorage;

        private const int NumRetries = 3;
        private static RetryPolicy _tableRetry = ProviderRetryPolicies.RetryN(NumRetries, TimeSpan.FromSeconds(1));

        #endregion


        #region Public methods

        /// <summary>
        /// This is the initialization function that initializes the member
        /// variables. It also creates the Photo table in the table storage
        /// if the table has not been created. This function is only called 
        /// once in the Application_Start function in Global.asax file.
        /// </summary>
        public static void Initialize()
        {
            StorageAccountInfo blobAccount = StorageAccountInfo.GetDefaultBlobStorageAccountFromConfiguration();
            _accountName = blobAccount.AccountName;
            _accountSharedKey = blobAccount.Base64Key;
            _storageEndpoint = blobAccount.BaseUri.AbsolutePath;
            _usePathStyleUris = blobAccount.UsePathStyleUris;

            _tableName = ConfigurationSettings.AppSettings["photoTableName"];
            _largePhotoBlobName = ConfigurationSettings.AppSettings["largePhotoBlobName"];
            _mediumPhotoBlobName = ConfigurationSettings.AppSettings["mediumPhotoBlobName"];
            _smallPhotoBlobName = ConfigurationSettings.AppSettings["smallPhotoBlobName"];
            _originalPhotoBlobName = ConfigurationSettings.AppSettings["originalPhotoBlobName"];


            try
            {
                _account = StorageAccountInfo.GetDefaultTableStorageAccountFromConfiguration();
                _tableStorage = TableStorage.Create(_account);

                _blobStorage = BlobStorage.Create(blobAccount);
                _blobStorage.RetryPolicy = RetryPolicies.RetryN(NumRetries, TimeSpan.FromMilliseconds(100));

                _tableStorage.TryCreateTable(_tableName);
            }
            catch (Exception e)
            {
                Log.Write(EventKind.Error, "Initialization of data service structures (tables and/or blobs) for photo storage failed!");
                throw new Exception("Could not create or find photo table and/or photo blob storage in the data service ", e);
            }
        }


        /// <summary>
        /// Add a photo in the Photo table and storage four different sized
        /// copies of the photo streams in the blob storage for this photo.
        /// </summary>
        /// <param name="albumId">The album Id</param>
        /// <param name="caption">The caption of the photo</param>
        /// <param name="bytesOriginal">The original-sized photo stream</param>
        /// <param name="bytesLarge">The large-sized photo stream</param>
        /// <param name="bytesMedium">The medium-sized photo stream</param>
        /// <param name="bytesSmall">The small sized photo stream</param>
        public static void AddPhoto(string albumId, string caption, byte[] bytesOriginal, byte[] bytesLarge, byte[] bytesMedium, byte[] bytesSmall)
        {
            SecUtility.CheckParameter(ref caption, false, false, false, TableStorageConstants.MaxStringPropertySizeInChars, "caption");
            string photoId = Guid.NewGuid().ToString();

            // Create a row in the Photo table for this photo
            PhotoRow newPhotoRow = new PhotoRow(albumId, photoId);
            newPhotoRow.Caption = caption;
            newPhotoRow.OriginalPhotoBlobName = photoId + _originalPhotoBlobName;
            newPhotoRow.LargePhotoBlobName = photoId + _largePhotoBlobName;
            newPhotoRow.MediumPhotoBlobName = photoId + _mediumPhotoBlobName;
            newPhotoRow.SmallPhotoBlobName = photoId + _smallPhotoBlobName;

            try
            {
                // Add the four different sized photo files to the blob storage before add the newPhotoRow
                // to the photo table in the table storage.
                AddPhotoFilesToBlobStorage(newPhotoRow, bytesOriginal, bytesLarge, bytesMedium, bytesSmall);

                // Add the newPhotoRow to the photo table in the table storage.
                TableStorageDataServiceContext svc = CreateDataServiceContext();
                svc.AddObject(_tableName, newPhotoRow);
                svc.SaveChanges();
            }
            catch (InvalidOperationException e)
            {
                HttpStatusCode status;
                // when retry policies are used we cannot distinguish between a conflict and success
                // so, in the case of a conflict, we just retrun success here
                if (TableStorageHelpers.EvaluateException(e, out status) && status == HttpStatusCode.Conflict)
                {
                    return;
                }

                Log.Write(EventKind.Warning, "Could not add the photo (ID={0}) to the Photo table.", photoId);
                throw new Exception("Failed to access the Photo table ", e);
            }
        }


        /// <summary>
        /// Retrieves a list of photos in the Photo table that belong to an album.
        /// </summary>
        /// <param name="albumId">The album Id</param>
        /// <returns>Returns a list of PhotoRow objects</returns>
        public static List<PhotoRow> GetPhotos(string albumId)
        {
            TableStorageDataServiceContext svc = CreateDataServiceContext();

            IEnumerable<PhotoRow> query = from c in svc.CreateQuery<PhotoRow>(_tableName)
                                           where c.PartitionKey == albumId
                                           select c;

            TableStorageDataServiceQuery<PhotoRow> q = new TableStorageDataServiceQuery<PhotoRow>(query as DataServiceQuery<PhotoRow>, _tableRetry);
            IEnumerable<PhotoRow> photoRows = q.ExecuteAllWithRetries();

            if (photoRows != null)
            {
                List<PhotoRow> list = new List<PhotoRow>(photoRows);
                return list;
            }

            return null;
        }


        /// <summary>
        /// Retrieves a list of photos in the Photo table that belong to an album.
        /// The album must be a public album or the filterPrivate is set to false.
        /// Otherwise, an empty list is returned.
        /// </summary>
        /// <param name="albumId">The album Id</param>
        /// <param name="filterPrivate">If filterPrivate is set to true, only
        /// photos belonging to a pulic album can be retrieved. Otherwise, photos
        /// belonging to either a public or non-public album can be retrieved.
        /// </param>
        /// <returns>Returns a list of PhotoRow objects</returns>
        public static List<PhotoRow> GetPhotos(string albumId, bool filterPrivate)
        {
            AlbumRow album = AlbumTable.GetAlbum(albumId);

            if (album != null && (album.IsPublic || !filterPrivate))
            {
                TableStorageDataServiceContext svc = CreateDataServiceContext();

                IEnumerable<PhotoRow> query = from c in svc.CreateQuery<PhotoRow>(_tableName)
                                              where c.PartitionKey == albumId
                                              select c;

                TableStorageDataServiceQuery<PhotoRow> q = new TableStorageDataServiceQuery<PhotoRow>(query as DataServiceQuery<PhotoRow>, _tableRetry);
                IEnumerable<PhotoRow> photoRows = q.ExecuteAllWithRetries();

                if (photoRows != null)
                {
                    List<PhotoRow> list = new List<PhotoRow>(photoRows);
                    return list;
                }
            }

            return null;
        }


        /// <summary>
        /// Retrieve the first photo in an album that is public or the  
        /// filterPrivate flag is set to false.
        /// </summary>
        /// <param name="albumId">The album Id</param>
        /// <param name="size">The type of the photo to be retrieved</param>
        /// <param name="filterPrivate"></param>
        /// <returns></returns>
        public static Stream GetFirstPhoto(string albumId, PhotoSize size, bool filterPrivate)
        {
            List<PhotoRow> photos = GetPhotos(albumId, filterPrivate);

            if (photos.Count > 0)
            {
                return GetPhoto(photos[0].PhotoId, size);
            }

            return null;
        }


        /// <summary>
        /// Retrieve a photo stream from the blob storage.
        /// </summary>
        /// <param name="photoId">The photo Id</param>
        /// <param name="size">The type of the photo to be retrieved</param>
        /// <param name="filterPrivate">If filterPrivate is set to true, only
        /// a public photo can be retrieved. Otherwise, both public and 
        /// non-public photo can be retrieved.
        /// </param>     
        /// <returns>Returns a byte stream of the photo requested</returns>
        public static Stream GetPhoto(string photoId, PhotoSize size)
        {
            try
            {
                TableStorageDataServiceContext svc = CreateDataServiceContext();

                IEnumerable<PhotoRow> query = from c in svc.CreateQuery<PhotoRow>(_tableName)
                                              where c.RowKey == photoId
                                              select c;

                TableStorageDataServiceQuery<PhotoRow> q = new TableStorageDataServiceQuery<PhotoRow>(query as DataServiceQuery<PhotoRow>, _tableRetry);
                IEnumerable<PhotoRow> photoRows = q.ExecuteWithRetries();

                int count = 0;
                foreach (PhotoRow photo in photoRows)
                {
                    if (count++ > 1)
                    {
                        Log.Write(EventKind.Warning, "There are multiple photos with the same ID (ID={0}) in the Photo table. Photo ID should be unique.", photoId);
                    }

                    return GetPhotoFileFromBlobStorage(photo, size);
                }
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "Could not retrieve the photo (ID={0}) in the Photo table.", photoId);
                throw new Exception("Failed to access the Photo table ", e);
            }

            return null;
        }


        /// <summary>
        /// Retrieve a photo stream from the blob storage if the album containing
        /// this photo is public or the filterPrivate flag is set to false.
        /// </summary>
        /// <param name="photoId">The photo Id</param>
        /// <param name="size">The type of the photo to be retrieved</param>
        /// <param name="filterPrivate">If filterPrivate is set to true, only
        /// a public photo can be retrieved. Otherwise, both public and 
        /// non-public photo can be retrieved.
        /// </param>
        /// <returns>Returns a byte stream of the photo requested</returns>
        public static Stream GetPhoto(string photoId, PhotoSize size, bool filterPrivate)
        {
            try
            {
                TableStorageDataServiceContext svc = CreateDataServiceContext();

                IEnumerable<PhotoRow> query = from c in svc.CreateQuery<PhotoRow>(_tableName)
                                              where c.RowKey == photoId
                                              select c;

                TableStorageDataServiceQuery<PhotoRow> q = new TableStorageDataServiceQuery<PhotoRow>(query as DataServiceQuery<PhotoRow>, _tableRetry);
                IEnumerable<PhotoRow> photoRows = q.ExecuteWithRetries();

                int count = 0;
                foreach (PhotoRow photo in photoRows)
                {
                    if (count++ > 1)
                    {
                        Log.Write(EventKind.Warning, "There are multiple photos with the same ID (ID={0}) in the Photo table. Photo ID should be unique.", photoId);
                    }

                    AlbumRow album = AlbumTable.GetAlbum(photo.AlbumId);
                    if (album != null && (album.IsPublic || !filterPrivate))
                    {
                        return GetPhotoFileFromBlobStorage(photo, size);
                    }
                }

                return null;
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "Could not retrieve the photo (ID={0}) in the Photo table.", photoId);
                throw new Exception("Failed to access the Photo table ", e);
            }
        }
        

        /// <summary>
        /// Remove all the photos for an album in the Photo table and in the 
        /// blob storage.
        /// </summary>
        /// <param name="albumId"></param>
        public static void RemovePhotos(string albumId)
        {
            try
            {
                RemoveAlbumFromBlobStorage(albumId);

                TableStorageDataServiceContext svc = CreateDataServiceContext();

                IEnumerable<PhotoRow> query = from c in svc.CreateQuery<PhotoRow>(_tableName)
                                              where c.PartitionKey == albumId
                                              select c;

                TableStorageDataServiceQuery<PhotoRow> q = new TableStorageDataServiceQuery<PhotoRow>(query as DataServiceQuery<PhotoRow>, _tableRetry);
                IEnumerable<PhotoRow> photoRows = q.ExecuteAllWithRetries();

                foreach (PhotoRow photo in photoRows)
                {
                    svc.DeleteObject(photo);
                    svc.SaveChanges();
                }
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "The deletion of the photos in the album (ID={0}) from the data storage failed.", albumId);
                throw new Exception("Failed to delete all the photos in the album ", e);
            }
        }


        /// <summary>
        /// Remove a row in the Photo table.
        /// </summary>
        /// <param name="photoId">The photo Id</param>
        public static void RemovePhoto(string photoId)
        {
            try
            {
                TableStorageDataServiceContext svc = CreateDataServiceContext();

                IEnumerable<PhotoRow> query = from c in svc.CreateQuery<PhotoRow>(_tableName)
                                              where c.RowKey == photoId
                                              select c;

                TableStorageDataServiceQuery<PhotoRow> q = new TableStorageDataServiceQuery<PhotoRow>(query as DataServiceQuery<PhotoRow>, _tableRetry);
                IEnumerable<PhotoRow> photoRows = q.ExecuteWithRetries();

                int count = 0;
                foreach (PhotoRow photo in photoRows)
                {
                    if (count++ > 1)
                    {
                        Log.Write(EventKind.Warning, "There are multiple photos with the same ID (ID={0}) in the Photo table. Photo ID should be unique.", photoId);
                    }

                    RemovePhotoFilesFromBlobStorage(photo);

                    svc.DeleteObject(photo);
                    svc.SaveChanges();
                }
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "The deletion of the photo (ID={0}) from the data storage failed.", photoId);
                throw new Exception("Failed to delete the photo ", e);
            }
        }


        /// <summary>
        /// Modify a row in the Photo table.
        /// </summary>
        /// <param name="caption">The caption of a photo</param>
        /// <param name="photoId">The photo Id </param>
        public static void EditPhoto(string caption, string photoId) 
        {
            SecUtility.CheckParameter(ref caption, false, false, false, TableStorageConstants.MaxStringPropertySizeInChars, "caption");

            try
            {
                TableStorageDataServiceContext svc = CreateDataServiceContext();

                IEnumerable<PhotoRow> query = from c in svc.CreateQuery<PhotoRow>(_tableName)
                                              where c.RowKey == photoId
                                              select c;

                TableStorageDataServiceQuery<PhotoRow> q = new TableStorageDataServiceQuery<PhotoRow>(query as DataServiceQuery<PhotoRow>, _tableRetry);
                IEnumerable<PhotoRow> photoRows = q.ExecuteWithRetries();

                int count = 0;
                foreach (PhotoRow photo in photoRows)
                {
                    if (count++ > 1)
                    {
                        Log.Write(EventKind.Warning, "There are multiple photos with the same ID (ID={0}) in the Photo table. Photo ID should be unique.", photoId);
                    }
                    photo.Caption = caption;
                    svc.UpdateObject(photo);
                    svc.SaveChanges();
                }
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "Could not edit the photo (ID={0}) in the Photo table.", photoId);
                throw new Exception("Failed to access the Photo table ", e);
            }
        }

        #endregion

        #region Helper methods
        private static TableStorageDataServiceContext CreateDataServiceContext()
        {
            return _tableStorage.GetDataServiceContext();
        }

        private static void AddPhotoFilesToBlobStorage(PhotoRow photo, 
                                                       byte[] bytesOriginal, 
                                                       byte[] bytesLarge, 
                                                       byte[] bytesMedium, 
                                                       byte[] bytesSmall)
        {
            try
            {
                // Create a blob for each different sized photo (original, large, medium and small) 
                // under the containerName=AlbumId in the blob storage.
                BlobContainer container = _blobStorage.GetBlobContainer(photo.AlbumId);

                NameValueCollection containerMetadata = new NameValueCollection();
                containerMetadata.Add("Name", "PersonalWebSiteAlbumContainer");
                container.CreateContainer(containerMetadata, ContainerAccessControl.Private);

                string blobName = photo.PhotoId + _originalPhotoBlobName;
                MemoryStream blobContent = new MemoryStream(bytesOriginal);
                BlobProperties properties = new BlobProperties(blobName);
                container.CreateBlob(properties, new BlobContents(blobContent), true);

                blobName = photo.PhotoId + _largePhotoBlobName;
                blobContent = new MemoryStream(bytesLarge);
                properties = new BlobProperties(blobName);
                container.CreateBlob(properties, new BlobContents(blobContent), true);

                blobName = photo.PhotoId + _mediumPhotoBlobName;
                blobContent = new MemoryStream(bytesMedium);
                properties = new BlobProperties(blobName);
                container.CreateBlob(properties, new BlobContents(blobContent), true);

                blobName = photo.PhotoId + _smallPhotoBlobName;
                blobContent = new MemoryStream(bytesSmall);
                properties = new BlobProperties(blobName);
                container.CreateBlob(properties, new BlobContents(blobContent), true);
            }
            catch (StorageException se)
            {
                Log.Write(EventKind.Warning, "Could not add the photo stream files for the photo (ID={0}) to the blob storage.", photo.PhotoId);
                throw new Exception("Failed to access blob storage ", se);
            }
        }

        private static Stream GetPhotoFileFromBlobStorage(PhotoRow photo, PhotoSize size)
        {
            string photoFileName;
            switch (size)
            {
                case PhotoSize.Small:
                    photoFileName = photo.SmallPhotoBlobName;
                    break;

                case PhotoSize.Medium:
                    photoFileName = photo.MediumPhotoBlobName;                            
                    break;

                case PhotoSize.Large:
                    photoFileName = photo.LargePhotoBlobName;
                    break;

                case PhotoSize.Original:
                default:
                    photoFileName = photo.OriginalPhotoBlobName;
                    break;
            }

            BlobContainer container = _blobStorage.GetBlobContainer(photo.AlbumId);
            try
            {
                MemoryStream outputStream = new MemoryStream();
                container.GetBlob(photoFileName, new BlobContents(outputStream), false);
                outputStream.Seek(0, SeekOrigin.Begin);
                return outputStream;
            }
            catch (StorageClientException se)
            {
                Log.Write(EventKind.Warning, "Could not retrieve the photo stream files for the photo (ID={0}) from the blob storage.", photo.PhotoId);
                throw new Exception("Failed to access blob storage ", se);
            }
        }

        private static void RemovePhotoFilesFromBlobStorage(PhotoRow photo)
        {
            BlobContainer container = _blobStorage.GetBlobContainer(photo.AlbumId);
            try
            {
                container.DeleteBlob(photo.OriginalPhotoBlobName);
                container.DeleteBlob(photo.SmallPhotoBlobName);
                container.DeleteBlob(photo.MediumPhotoBlobName);
                container.DeleteBlob(photo.LargePhotoBlobName);
            }
            catch (StorageException se)
            {
                Log.Write(EventKind.Warning, "The deletion of photo files for the photo (ID={0}) from the blob storage failed.", photo.PhotoId);
                throw new Exception("Failed to access blob storage ", se);
            }
        }

        private static void RemoveAlbumFromBlobStorage(string albumId)
        {
            BlobContainer container = _blobStorage.GetBlobContainer(albumId);
            try
            {
                if (container.DoesContainerExist())
                {
                    container.DeleteContainer();
                }
            }
            catch (StorageClientException se)
            {
                Log.Write(EventKind.Warning, "The deletion of all the photos belongung to the album (ID={0}) from the blob storage failed.", albumId);
                throw new Exception("Failed to access blob storage ", se);
            }
        }
        #endregion
    }

}
